unit unzip;   {Unzips deflated, imploded and stored files}

{$C moveable demandload discardable}
{$ifdef win32}
{$A-}
{$O-}
{$endif}
{C code by info-zip group, translated to pascal by Christian Ghisler}
{based on unz51g.zip}
{Special thanks go to Mark Adler,
 who wrote the main inflate and explode code,
 and did NOT copyright it!!!}

interface

{$define assembler}      {use assembler!}

{Please note: do NOT add $define wincmd. This flag is used for the integration of the unzip
 unit in the author's own program. It does 2 things: load strings for the swap disk option
 from resource, and update a directory tree when a new directory is created}

{$R-}         {No range checking}

{$ifdef win32}
{$define windows}
{$endif}

{$ifdef windows}
uses messages,wintypes,winprocs,sysutils,packdefs;

{$else}
uses strings,dos,crt;
{$endif}

{$ifndef wincmd}
const
  cryptkey1:longint=-1;              {Keys for ZIP-Encryption}
  cryptkey2:longint=-1;
  cryptkey3:longint=-1;
{$endif}

const   {Error codes returned by unzip}
  unzip_Ok=0;
  unzip_CRCErr=1;
  unzip_WriteErr=2;
  unzip_ReadErr=3;
  unzip_ZipFileErr=4;
  unzip_UserAbort=5;
  unzip_NotSupported=6;
  unzip_Encrypted=7;
  unzip_InUse=-1;

function GetSupportedMethods:longint;
{Checks which pack methods are supported by the dll}
{bit 8=1 -> Format 8 supported, etc.}

function unzipfile(in_name:pchar;out_name:pchar;attr:word;offset,originalsize:comp;var hFileAction:thandle;
  cm_index:integer;var unpackhandle:file):integer;
{usage:
 in_name:      name of zip file with full path
 out_name:     desired name for out file
 offset:       header position of desired file in zipfile
 originalsize: size of file from central directory, needed only for stored files in stream mode
 hFileAction:  handle to dialog box showing advance of decompression (optional)
 cm_index:     notification code sent in a wm_command message to the dialog
               to update percent-bar
 Return value: one of the above unzip_xxx codes

 Example for handling the cm_index message in a progress dialog:

 unzipfile(......,cm_showpercent);

 ...

 procedure TFileActionDialog.wmcommand(var msg:tmessage);
 var ppercent:^word;
 begin
   TDialog.WMCommand(msg);
   if msg.wparam=cm_showpercent then begin
     ppercent:=pointer(lparam);
     if ppercent<>nil then begin
       if (ppercent^>=0) and (ppercent^<=100) then
         SetProgressBar(ppercent^);
       if UserPressedAbort then
         ppercent^:=$ffff
       else
         ppercent^:=0;
       end;
     end;
   end;
 end;
}

function unzipfiletomemory(in_name:pchar;out_buf:pchar;var buf_size:longint;
  offset,originalsize:comp;var hFileAction:thandle;cm_index:integer;var unpackhandle:file):integer;
{usage:
 in_name:      name of zip file with full path
 out_buf:      buffer to recieve unpacked file
 buf_size:     size of buffer to recieve unpacked file
 offset:       header position of desired file in zipfile
 hFileAction:  handle to dialog box showing advance of decompression (optional)
 cm_index:     notification code sent in a wm_command message to the dialog
               to update percent-bar
 Return value: one of the above unzip_xxx codes
}

function UnzipTestIntegrity(in_name:pchar;offset,originalsize:comp;
  var hFileAction:thandle;cm_index:integer;var crc:longint;var unpackhandle:file):integer;

function UnzipTestPassword(in_name:pchar;offset:comp;
  var hFileAction:thandle;cm_index:integer;var crc:longint;var unpackhandle:file):integer;

procedure CloseZipFileHandle(var unpackhandle:file);
{Close handle after unpacking one or more files from zip}

function asktoswapdisk(parent:hwnd;newdisk:cardinal;newname:pchar):boolean;   {Ask user to swap disks (multi-volume)}

function UnzipSetPassword(password:pchar):boolean;

function IsZipPasswordSet:boolean;
function ziplastwasencrypted:boolean;

implementation

{$ifdef wincmd}
uses filedlgs,filestuf{$ifdef win32},pkplugin,makezip{$endif};
{$endif}  {For movetoback}

{$ifndef win32}
type short=integer;
var currentdisk:word;
{$else}
threadvar currentdisk:integer;       {current disk in drive}
{$endif}

const specialmaxcomp:comp=100.0*maxlongint;

function asktoswapdisk(parent:hwnd;newdisk:cardinal;newname:pchar):boolean;   {Ask user to swap disks (multi-volume)}
var buf:array[0..79] of char;
    newname2:tdirtype;
    p:pchar;
    bufn:array[0..3] of char;
    matched:boolean;
begin
  asktoswapdisk:=true;
  if newdisk=$FFFF then currentdisk:=$FFFF   {Init}
  else if newdisk=$FFFE then currentdisk:=$FFFE   {Init: not multivol!}
  else if (newdisk<=0) or (newdisk>1000) then    {Possible error}
    asktoswapdisk:=false
  else if (newdisk<>currentdisk) and (currentdisk<>$FFFE) then begin
    asktoswapdisk:=false;
    {$ifdef wincmd}
    {$ifdef win32}
    wvsprintf(buf,getstring(s_insertdiski),pchar(@newdisk));
    {$else}
    wvsprintf(buf,getstring(s_insertdiski),newdisk);
    {$endif}
    {$else}
    {$ifdef win32}
    wvsprintf(buf,'Please insert disk number %i and press ENTER!',pchar(@newdisk));
    {$else}
    wvsprintf(buf,'Please insert disk number %i and press ENTER!',newdisk);
    {$endif}
    {$endif}
    {$ifdef Windows}
    {$ifdef wincmd}
    if (newname<>nil) and (newdisk<1000) then begin
      {Try to access Winzip-style split zip files!}
      {Naming: .z01, .z02, .z03 ... .zip}
      {Note: The LAST file is named .zip!}
      strlcopy(newname2,newname,sizeof(newname2)-1);
      p:=strrscan(newname2,'.');
      if p<>nil then begin
        p[1]:=#0;
        strcopy(bufn,'z00');
        if newdisk>=100 then
          bufn[0]:=char(newdisk div 100+ord('0'));
        bufn[1]:=char((newdisk div 10) mod 10+ord('0'));
        bufn[2]:=char(newdisk mod 10+ord('0'));
        strlcat(newname2,bufn,sizeof(newname2)-1);
        if fileexists(strpas(newname2)) then begin
          currentdisk:=newdisk;
          result:=true;
          strlcopy(newname,newname2,sizeof(tdirtype)-1);
          exit;
        end;
        p[1]:=#0;
        strlcat(newname2,'zip',sizeof(newname2)-1);
        matched:=fileexists(strpas(newname2));
        if matched then begin
          matched:=matched and (stricomp(newname2,newname)<>0); {Was different?}
          if not matched then begin {No? -> check whether zipname.001 is present!}
            p[1]:=#0;
            strlcat(newname2,'z01',sizeof(newname2)-1);
            matched:=fileexists(strpas(newname2));
            p[1]:=#0;
            strlcat(newname2,'zip',sizeof(newname2)-1);
          end;
        end;
        if matched then begin
          currentdisk:=newdisk;
          result:=true;
          strlcopy(newname,newname2,sizeof(tdirtype)-1);
          exit;
        end;
      end;
      result:=InputdlgThread(parent,'Unzip',buf,newname,sizeof(tdirtype)-1,true,false,false)=idOk;
      if result then
        currentdisk:=newdisk;
    end else{$endif} begin
      messagebeep(0);
      if messagebox(parent,buf,'Unzip',mb_okcancel or mb_iconexclamation)=idok then begin
        result:=true;
        currentdisk:=newdisk;
      end;
    end;
    {$else}
    write(buf);
    readln;
    currentdisk:=newdisk;
    result:=true;
    {$endif}
  end;
end;

{$I z_tables.pas}  {Tables for bit masking, huffman codes and CRC checking}
{$I z_crc.pas}
{$I z_types.pas}   {global types and constants}

{$ifdef win32}
threadvar DefKey:array[0..2] of longint;
          aes_password:tdirtype;
          keythere:bool;
          ziplastcrypted:bool;
{$else}
var DefKey:array[0..2] of longint;
    keythere:bool;
{$endif}

procedure update_defkeys(ch:char);
var ch2:byte;
begin
  UpdateCRC(@ch,1,Defkey[0]);
  DefKey[1]:=DefKey[1] + (DefKey[0] and $000000ff);
  DefKey[1]:=DefKey[1]*134775813 + 1;
  ch2:=DefKey[1] shr 24;
  UpdateCRC(@ch2,1,DefKey[2]);
end;

function UnzipSetPassword(password:pchar):boolean;
var i:integer;
begin
  keythere:=false;
  ziplastcrypted:=false;
  if password<>nil then begin
    {$ifdef win32}
    strlcopy(aes_password,password,sizeof(aes_password)-1);
    {$endif}
    DefKey[0]:=305419896;
    DefKey[1]:=591751049;
    DefKey[2]:=878082192;
    for i:=0 to strlen(password)-1 do
      update_defkeys(password[i]);
    keythere:=password[0]<>#0;
    result:=true;
  end else begin
    fillchar(aes_password,sizeof(aes_password),#0);
    keythere:=false;
    result:=false;
  end;
end;

function IsZipPasswordSet:boolean;
begin
  result:=keythere;
end;

function ziplastwasencrypted:boolean;
begin
  result:=ziplastcrypted;
end;

{*************************************************************************}
{$ifndef win32}

var Pcplens,Pcpdist,Pcplext,Pcpdext:pushlist;    {Need to be adjusted for deflate or deflate64!}
    deflate64mode:boolean;

{$ifdef encryption}
procedure DecryptBuffer(indata:pchar;datalength:integer); forward;
{$endif}
{$I z_global.pas}  {global variables}
{$I z_generl.pas}  {General functions used by both inflate and explode}
{$I z_huft.pas}    {Huffman tree generating and destroying}
{$ifdef encryption}
{$I z_crypt.pas}   {Decrypt}
{$endif}
{$I z_inflat.pas}  {Inflate deflated file}
{$I z_copyst.pas}  {Copy stored file}
{$ifndef installer}
{$I z_explod.pas}  {Explode imploded file}
{$I z_shrunk.pas}  {Unshrink function}
{$I z_expand.pas}  {Expand functions}
{$endif}
{$endif}

{$ifdef win32}
{$ifndef installer}
threadvar bz2_oldtick:longint;
          bz2_oldpercent:integer;
          bz2_dlghandle:thandle;
          bz2_dlgnotify:integer;
          bzip2_key0,bzip2_key1,bzip2_key2:longint;
          bzip2_aes_context,bzip2_aes_context_backup:longint;

function bzcallback(percent:integer):bool; stdcall;
begin
  result:=false;
  if percent>100 then percent:=100;
  if (abs(gettickcount-bz2_oldtick)>=100) then begin
    bz2_oldtick:=gettickcount;
    bz2_oldpercent:=percent;
    if bz2_dlghandle<>0 then begin     {Use dialog box for aborting}
      {Sendmessage returns directly -> ppercent contains result}
      sendmessage(thandle(bz2_dlghandle),wm_command,bz2_dlgnotify,longint(@percent));
      result:=(percent=$FFFF);   {Abort pressed!}
    end else if bz2_dlgnotify<>0 then
      result:=getasynckeystate(bz2_dlgnotify)<0;  {break Key pressed!}
    {$ifdef win32}
    {$ifdef wincmd}
    if (bz2_dlghandle<>0) and TestIfBkGndClicked(thandle(bz2_dlghandle)) then
      MoveToBack(thandle(bz2_dlghandle));
    {$endif}
    {$endif}
  end;
end;

function requestnewvol(newfilename:pchar):bool; stdcall;
begin
  result:=asktoswapdisk(bz2_dlghandle,currentdisk+1,newfilename);
end;

{ZIP encryption}
function unscramblecallback1(buffer:pchar;len:integer):bool; stdcall;
var local_Key0,local_Key1,local_Key2:longint;

  procedure update_keys(ch:byte);
  var ch2:byte;
  begin
    UpdateCRC(@ch,1,local_key0);
    local_Key1:=local_Key1 + (local_Key0 and $000000ff);
    local_Key1:=local_Key1*134775813 + 1;
    ch2:=local_key1 shr 24;
    UpdateCRC(@ch2,1,local_key2);
  end;

  function decrypt_byte:byte;
  var temp:longint;
  begin
    temp:=(local_Key2 and $FFFF) or 2;
    decrypt_byte:=(temp * (temp xor 1)) shr 8;
  end;

  var i:integer;
      C:byte;

begin
  local_Key0:=bzip2_key0;
  local_Key1:=bzip2_key1;
  local_Key2:=bzip2_key2;
  for i:=0 to len-1 do begin
    C:=byte(buffer[i]) xor decrypt_byte;
    update_keys(C);
    buffer[i]:=char(C);
  end;
  bzip2_key0:=local_Key0;
  bzip2_key1:=local_Key1;
  bzip2_key2:=local_Key2;
  result:=true;
end;

{AES encryption}
function unscramblecallback2(buffer:pchar;len:integer):bool; stdcall;
var aes_context_ptr,aes_context_backup_ptr:pfcrypt_ctx;
begin
  aes_context_ptr:=pfcrypt_ctx(bzip2_aes_context);
  aes_context_backup_ptr:=pfcrypt_ctx(bzip2_aes_context_backup);
  if len>=0 then
    aes_context_backup_ptr^:=aes_context_ptr^      // save context _before_ crypt, for multi-volume
  else begin
    aes_context_ptr^:=aes_context_backup_ptr^;     // restore context, we went too far
  end;
  //ZIP_DecryptZipAES(buffer,abs(len),aes_context_ptr^);
  result:=true;
end;
{$endif}
{$endif}

type grec=record
       infile:file;        {handle to zipfile}
       outfile:file;       {handle to extracted file}
       infilename:array[0..259] of char;  {file name of input file}
       compsize,               {comressed size of file}
       uncompsize,             {uncompressed size of file}
       reachedsize:comp;       {number of bytes read from zipfile}
       crc32val:longint;       {crc calculated from data}
       hufttype:word;          {coding type=bit_flag from header}
       dlghandle:thandle;      {optional: handle of a cancel and "%-done"-dialog}
       dlgnotify:short;        {notification code to tell dialog how far the decompression is}
       aes_strength:byte;      {0=no AES, 1=128 bit, 2=192 bit, 3=256 bit}
     end;

function DoRealUnpack(unpackzip:boolean;ziptype:integer;var originalcrc:longint;offset:comp;
  var gr:grec;var g_byteswritten:longint;g_tomemory,g_crconly:boolean;g_outbuf:pointer;
      g_bufsize:longint;g_testpassword:boolean):integer;
{$ifdef win32}
{$ifdef encryption}
procedure DecryptBuffer(indata:pchar;datalength:integer); forward;
{$ifndef installer}
procedure AesDecryptBuffer(indata:pchar;datalength:integer); forward;
{$endif}
{$endif}
var wsize:integer;
{$I z_global.pas}  {global constants, types and variables}
{$I z_generl.pas}  {General functions used by both inflate and explode}
{$I z_huft.pas}    {Huffman tree generating and destroying}
{$ifdef encryption}
{$I z_crypt.pas}   {Decrypt}
{$endif}

var Pcplens,Pcpdist,Pcplext,Pcpdext:pushlist;    {Need to be adjusted for deflate or deflate64!}

{$I z_inflat.pas}  {Inflate deflated file}
{$I z_copyst.pas}  {Copy stored file}
{$ifndef installer}
{$I z_explod.pas}  {Explode imploded file}
{$I z_shrunk.pas}  {Unshrink function}
{$I z_expand.pas}  {Expand functions}
{$endif}
{$endif}
{$ifndef win32}
var slidemem:thandle;
{$else}
var res:integer;
    enddatablock:array[0..1] of longint;
{$ifdef encryption}
    aes_pwd_ver:pwdver_array;
    unscramblecallback:tunscramblecallbackproc;
    saltlen:integer;
{$endif}
{$endif}
var readl,sig1,sig2:word;
begin
  {$ifdef win32}
  if ziptype=9 then wsize:=wsize64
               else wsize:=wsize32;
  {$else}
  deflate64mode:=ziptype=9;
  {16-bit version: patch the tables directly to conserve memory}
  if deflate64mode then begin
    cplens[28]:=3;
    cplext[28]:=16;
    cpdext[30]:=14;
    cpdext[31]:=14;
  end else begin
    cplens[28]:=258;
    cplext[28]:=0;
    cpdext[30]:=99;
    cpdext[31]:=99;
  end;
  {$endif}

  oldtick:=-1;   {force first refresh!}
  global_byteswritten:=g_byteswritten;
  global_tomemory:=g_tomemory;
  global_crconly:=g_crconly;
  global_outbuf:=g_outbuf;
  global_bufsize:=g_bufsize;
  infile:=gr.infile;
  outfile:=gr.outfile;
  strcopy(infilename,gr.infilename);
  compsize:=gr.compsize;
  uncompsize:=gr.uncompsize;
  reachedsize:=gr.reachedsize;
  crc32val:=gr.crc32val;
  hufttype:=gr.hufttype;
  dlghandle:=gr.dlghandle;
  dlgnotify:=gr.dlgnotify;
  {$ifdef win32}
  global_aes_strength:=gr.aes_strength;
  {$endif}
  totalabort:=false;
  crc32val:=$FFFFFFFF;

  {$ifdef encryption}
  decrypt:=false;
  {$endif}
  {$ifndef win32}
  if deflate64mode then begin
    slidemem:=globalalloc(GMEM_MOVEABLE or GMEM_ZEROINIT,$10000);
    slide:=GlobalLock(slidemem);
  end else
  {$endif}
  begin
    getmem(slide,wsize);
    fillchar(slide[0],wsize,#0);
  end;
  zipeof:=false;
  reachedsize:=0;
  messageloop;
  {$ifdef windows}
  oldpercent:=0;
  if dlghandle<>0 then
    sendmessage(dlghandle,wm_command,dlgnotify,longint(@oldpercent));  {0 Percent}
  {$endif}
  if not unpackzip then begin  {GZIP archive, not ZIP}
    Pcplens:=pushlist(@cplens);
    Pcplext:=pushlist(@cplext);
    Pcpdext:=pushlist(@cpdext);
    Pcpdist:=pushlist(@cpdist);
    Result:=inflate(0);
    compsize:=compsize+4;{Avoid reachedsize>compsize+2 in readbuf, because reading over end of compressed data!}
    dumpbits(k and 7);
    needbits(16);
    originalcrc:=b and $FFFF;
    dumpbits(16);
    needbits(16);
    originalcrc:=originalcrc or (b and $FFFF) shl 16;
    dumpbits(16);
    needbits(16); {Flush the size!}
    dumpbits(16);
    needbits(16);
    dumpbits(16);
    compsize:=compsize-4;
    offset:=offset+reachedsize-readpos+inpos-1;
    if k>=8 then offset:=offset-1;  {Grund: needbits liest immer 2 bytes auf einmal!!!}
    reachedsize:=0;
    CompSeek(infile,offset,0);
  end else begin
    {$ifdef encryption}
    if (gr.hufttype and 1)<>0 then begin  {encrypted}
      {$ifdef win32}
      {$ifndef installer}
      if (global_aes_strength>=1) and (global_aes_strength<=3) then begin
        saltlen:=salt_size[global_aes_strength];
        fillchar(slide[0],saltlen+2,#0);
        readplaintextblock(slide,saltlen+2);  {Use slide as buffer, supports multi-volume}
        if (compsize=0) and (ziptype=0) then
          compsize:=uncompsize      {WinZIP-Multi-Volume stored!}
        else
          compsize:=compsize-saltlen-2{Password verification value}-10{auth};
        fillchar(aes_context,sizeof(aes_context),#0);

        if gr.compsize=specialmaxcomp then begin    {Can't check if size unknown!}
          compsize:=400000000;
          compsize:=compsize*1000;
        end;
        res:=ZIP_InitZipAES(global_aes_strength,aes_password,strlen(aes_password),slide,
          aes_pwd_ver,aes_context);
        if (res<>0) then
          ziptype:=99
        else if (aes_pwd_ver[0]<>byte(slide[saltlen])) or (aes_pwd_ver[1]<>byte(slide[saltlen+1])) then
          ziptype:=-1;
      end else
      {$endif}
      {$endif}
      begin
        decrypt:=true;
        SetDefKey;
      end;
    end;

    if g_testpassword then begin
      inpos:=0;            {Input buffer position}
      readpos:=-1;         {Nothing read}
      w:=0;                {initialize window, bit buffer}
      k:=0;
      b:=0;
      if ((gr.hufttype and 1)<>0) and
        ((global_aes_strength>=1) and (ziptype<>99) and (ziptype<>-1) or
         (global_aes_strength=0) and CheckCryptCrc(originalcrc)) then begin
        gr.crc32Val:=originalcrc;
        result:=0
      end else begin
        gr.crc32Val:=originalcrc+1;
        result:=unzip_crcerr;
      end;
      {$ifndef win32}
      if deflate64mode then begin
        GlobalUnLock(slidemem);
        GlobalFree(slidemem);
      end else
      {$endif}
      freemem(slide,wsize);
      exit;
    end;
    {$endif}
    if (compsize=0) and (ziptype=0) then
      compsize:=uncompsize;      {WinZIP-Multi-Volume stored!}
    {Unzip correct type}
    if (outfile<>invalid_handle_value) and (global_aes_strength>=1) and (global_aes_strength<=3) and (ziptype<>-1) then
      longtruncate(outfile);

    case ziptype of
     -1:Result:=unzip_CRCErr;
      0:Result:=copystored(originalcrc);
    {$ifndef installer}
      1:Result:=unshrink(originalcrc);
      2..5:Result:=unreduce(ziptype,originalcrc);
      6:Result:=explode(originalcrc);
    {$endif}
      8:begin
          Pcplens:=pushlist(@cplens);
          Pcplext:=pushlist(@cplext);
          Pcpdext:=pushlist(@cpdext);
          Pcpdist:=pushlist(@cpdist);
          Result:=inflate(originalcrc);
        end;
      9:begin
    {$ifdef win32}
          Pcplens:=pushlist(@cplens64);  {Use separate tables for multi-threading!}
          Pcplext:=pushlist(@cplext64);
          Pcpdext:=pushlist(@cpdext64);
          Pcpdist:=pushlist(@cpdist);
    {$else}
          Pcplens:=pushlist(@cplens);    {Tables were patched above!}
          Pcplext:=pushlist(@cplext);
          Pcpdext:=pushlist(@cpdext);
          Pcpdist:=pushlist(@cpdist);
    {$endif}
          Result:=inflate(originalcrc);
        end;
    {$ifdef win32}
    {$ifndef installer}
      12:begin
           if LoadUnzipLibrary and (@UnzipBzip2<>nil) then begin
             bz2_dlghandle:=dlghandle;
             bz2_dlgnotify:=dlgnotify;
             unscramblecallback:=nil;
             {$ifdef encryption}
             if (gr.hufttype and 1)<>0 then begin  {encrypted}
               if (global_aes_strength>=1) then begin
                 bzip2_aes_context:=longint(@aes_context);
                 aes_context_backup:=aes_context;
                 bzip2_aes_context_backup:=longint(@aes_context_backup);
                 unscramblecallback:=unscramblecallback2;
               end else begin
                 readpos:=0;   {Clear upper 16 bit}
                 longblockread(infile,inbuf,12,word(readpos)); {just skip!}
                 DecryptBuffer(pchar(@inbuf),readpos);
                 bzip2_key0:=global_Key[0];
                 bzip2_key1:=global_Key[1];
                 bzip2_key2:=global_Key[2];
                 unscramblecallback:=unscramblecallback1;
               end;
             end;
             {$endif}
             Result:=UnzipBzip2(infilename,infile,outfile,crc32val,@bzcallback,unscramblecallback,
               @requestnewvol,compsize,uncompsize)
           end else
             Result:=unzip_NotSupported;
         end;
      98:begin
           if LoadUnzipLibrary and (@UnzipPpmd<>nil) then begin
             bz2_dlghandle:=dlghandle;
             bz2_dlgnotify:=dlgnotify;
             unscramblecallback:=nil;
             {$ifdef encryption}
             if (gr.hufttype and 1)<>0 then begin  {encrypted}
               if (global_aes_strength>=1) then begin
                 bzip2_aes_context:=longint(@aes_context);
                 aes_context_backup:=aes_context;
                 bzip2_aes_context_backup:=longint(@aes_context_backup);
                 unscramblecallback:=unscramblecallback2;
               end else begin
                 readpos:=0;   {Clear upper 16 bit}
                 longblockread(infile,inbuf,12,word(readpos)); {just skip!}
                 DecryptBuffer(pchar(@inbuf),readpos);
                 bzip2_key0:=global_Key[0];
                 bzip2_key1:=global_Key[1];
                 bzip2_key2:=global_Key[2];
                 unscramblecallback:=unscramblecallback1;
               end;
             end;
             {$endif}
             Result:=UnzipPpmd(infilename,infile,outfile,crc32val,@bzcallback,
                 unscramblecallback,@requestnewvol,compsize,uncompsize)
           end else
             Result:=unzip_NotSupported;
         end;
    {$endif}
    {$endif}
    else
      Result:=unzip_NotSupported;
    end;
    gr.infile:=infile;    {May have changed due to disk change!}

    g_byteswritten:=global_byteswritten;
    if (g_byteswritten<>uncompsize) and (outfile<>-1) and (ziptype<>-1) then
      LongTruncate(outfile);       {Reason: Buffer-preallocation on NTFS!}

    {$ifdef win32}
    {$ifdef encryption}
    if global_aes_strength<>0 then begin
      if (gr.compsize=specialmaxcomp) and (ziptype<>98) and (ziptype<>12) and (ziptype<>14) and
        (readpos<sizeof(inbuf)) then begin    {We need to re-read the last slide read and recalc the hash!}
        dumpbits(k and 7);
        aes_context:=aes_context_backup;
        longseek(infile,-readpos-1,FILE_CURRENT); // readpos is 1 too low!
        longblockread(infile,inbuf,inpos-(k div 8),word(readpos));  // k div 8 must be substracted, because we read too much!!!
        AesDecryptBuffer(pchar(@inbuf),readpos);
      end;
      res:=ZIP_EndZipAES(slide,aes_context); {Nur schreibender Zugriff auf slide!}
      if res=10 then begin
        readl:=readplaintextblock(slide+10,10);  {Use slide as buffer, supports multi-volume}
        //longblockread(infile,slide[10],10,readl);  {Use slide as buffer}
        crc32val:=not originalcrc;
        if (gr.compsize<>specialmaxcomp) then
          readl:=10;          {Test all bytes}
        for res:=0 to readl-1 do
          if slide[res]<>slide[res+10] then
            crc32val:=not originalcrc+1;
      end;
    end;
    {$endif}
    {$endif}
    if (Result=unzip_ok) and ((hufttype and 8)<>0)
      and ((originalcrc<>not(crc32val))) then begin {CRC at the end}
      if (ziptype<>0) and (ziptype<>12) and (ziptype<>98) then begin
        dumpbits(k and 7);
        {When file is encrypted, we need to read the data again UNENCRYPED from the file!}
        {Reason: The headers are always unencrypted!!!}
        if (gr.hufttype and 1)<>0 then begin  {encrypted}
          {We are at location inpos of inbuf. The file pointer is at the END of inbuf, length readpos!}
          longseek(infile,-readpos-1-(k div 8)+inpos,FILE_CURRENT);
          enddatablock[0]:=0;
          longblockread(infile,enddatablock[0],sizeof(enddatablock),readl);
          sig1:=enddatablock[0] and $FFFF;
          sig2:=enddatablock[0] shr 16;
          if (sig1<>ord('P')+256*ord('K')) or
             ((sig2<>7+256*8) and (sig2<>7)) then begin     {PK#7#8 or PK#7#0 not there}
            originalcrc:=enddatablock[0];
          end else
            originalcrc:=enddatablock[1];
        end else begin
          needbits(16);     {pk#7#8 not always there!!!}
          sig1:=b and $FFFF;
          dumpbits(16);
          needbits(16);
          sig2:=b and $FFFF;
          dumpbits(16);
          if (sig1<>ord('P')+256*ord('K')) or
             ((sig2<>7+256*8) and (sig2<>7)) then begin     {PK#7#8 or PK#7#0 not there}
            originalcrc:=sig1+65536*sig2;
          end else begin
            needbits(16);                   {pk#7#8 found}
            originalcrc:=b and $FFFF;
            dumpbits(16);
            needbits(16);
            originalcrc:=originalcrc or (b and $FFFF) shl 16;
            dumpbits(16);
          end;
        end;
      end else begin   {Stored!}
        longblockread(infile,slide[0],64,readl);  {Use slide as buffer}
        if readl=64 then
        (*  {$ifdef win32}
          {We can't check the AES checksum on multi-volume archives, so
           check whether the correct header comes!}
          if (global_aes_strength<>0) and (gr.compsize=specialmaxcomp) then begin
            if (slide[0]<>'P') or (slide[1]<>'K') or (slide[2]<>#7) or (slide[3]<>#8) then
              crc32val:=not originalcrc+1;
          end else
          {$endif}*)
            originalcrc:=plongint(@slide[4])^;
      end;
    end;
  end;
  gr.crc32val:=not(crc32val);  {one's complement}
  gr.dlghandle:=dlghandle;
  {$ifndef win32}
  if deflate64mode then begin
    GlobalUnLock(slidemem);
    GlobalFree(slidemem);
  end else
  {$endif}
  freemem(slide,wsize);
  if strcomp(infilename,gr.infilename)<>0 then {Winzip split-files!}
    strlcopy(gr.infilename,infilename,sizeof(tdirtype)-1);
end; {DoRealUnpack}

{***************************************************************************}

function GetSupportedMethods:longint;
begin
  GetSupportedMethods:=1+(1 shl 1)+(1 shl 2)+(1 shl 3)+(1 shl 4)+(1 shl 5)+(1 shl 6)+(1 shl 8)
    +(1 shl 9) {$ifdef win32} +(1 shl 12){$endif};
  {stored, shrunk, imploded and deflated, and new deflate64 and bzip2!}
end;

function makecomp(lowlong,highlong:longint):comp;
type tcomprec=record
       case boolean of
         true:(l,h:longint);
         false:(cmp:comp);
       end;
var tcr:tcomprec;
begin
  tcr.l:=lowlong;
  tcr.h:=highlong;
  result:=tcr.cmp;
end;

type p64localheader=^t64localheader;
     t64localheader=packed record
       Tag,
       Size:word;
       orglow,orghi:longint;
       complow,comphi:longint;
     end;
     paeslocalheader=^taeslocalheader;
     taeslocalheader=packed record
       Tag,
       Size:word;
       vendorversion:word;
       vendorid:array[0..1] of char;
       aes_strength:byte;
       zip_compression:word;
     end;

{******************** main function: unzipfile *****************************}
{written and not copyrighted by Christian Ghisler}

function UnzipFileInternal(in_name:pchar;out_name:pchar;attr:word;offset,originalsize:comp;
  var hFileAction:thandle;cm_index:integer;var g_byteswritten:longint;
  g_tomemory,g_crconly:boolean;g_outbuf:pointer;g_bufsize:longint;
  var resCRC:longint;var unpackhandle:file;g_testpassword:boolean):integer;
var err:word;
    header:tlocalheader;
    outname:array[0..259] of char;
    compoffset,createsize,gtemp:comp;
    timedate:longint;
    p64,extrafield:p64localheader;
    extrafield_aes:paeslocalheader;
    originalcrc:longint;    {crc from zip-header}
    finalpercent:short;
    ziptype,flag:integer;
    extrafieldread:word;
    p:pchar;
    isadir,allowdeletebad:boolean;
    g:grec;

begin
  ziplastcrypted:=false;
  if (out_name<>nil) and (stricomp(in_name,out_name)=0) then begin {Just additional check, main check in packstuf}
    result:=unzip_WriteErr;
    exit
  end;
  {$ifdef windows}
  {$ifndef win32}
  if inuse then begin
    {take care of crashed applications!}
    if (lastusedtime<>0) and
      (abs(gettickcount-lastusedtime)>30000) then begin {1/2 minute timeout!!!}
      {do not close files or free slide, they were already freed when application crashed!}
      inuse:=false;
      {memory for huffman trees is lost}
    end else begin
      result:=unzip_inuse;
      exit
    end;
  end;
  inuse:=true;
  {$endif}
  {$endif}
  strcopy(g.infilename,in_name);
  if unpackhandle<>INVALID_HANDLE_VALUE then begin
    g.infile:=unpackhandle;
    doserror:=0;
  end else
    g.infile:=longreset(g.infilename,0,0);
  if doserror<>0 then     {Read error -> Try again (for floppies!)}
    g.infile:=longreset(g.infilename,0,0);

  if doserror<>0 then begin
    unpackhandle:=INVALID_HANDLE_VALUE;
    result:=unzip_ReadErr;
    inuse:=false;
    exit
  end;
  unpackhandle:=g.infile;
  compoffset:=offset;{4 GB hack!}
  compseek(g.infile,compoffset,0);       {seek to header position}
  if doserror<>0 then begin
    result:=unzip_ZipFileErr;
    inuse:=false;
    exit
  end;
  longblockread(g.infile,header,sizeof(header),err);  {read in local header}
  if (doserror<>0) or (sizeof(header)<>err) then begin
    result:=unzip_ZipFileErr;
    inuse:=false;
    exit
  end;

  if (strlcomp(header.signature,'PK'#3#4,4)<>0) then begin
    result:=unzip_ZipFileErr;
    inuse:=false;
    exit
  end;

  {calculate offset of data}
  timedate:=header.file_timedate;
  g.hufttype:=header.bit_flag;
  g.compsize:=makecomp(header.compress_size,0);
  g.uncompsize:=makecomp(header.uncompress_size,0);
  g.aes_strength:=0;  {No aes encryption}
  ziptype:=header.zip_type;     {0=stored, 6=imploded, 8=deflated}
  {Get 64 bit file sizes if available!}
  if ((header.compress_size=-1) or (header.uncompress_size=-1) or (ziptype=99)) and
      (header.extra_field_len>=sizeof(taeslocalheader)) then begin
    compseek(g.infile,compoffset+sizeof(header)+header.filename_len,0);
    getmem(p64,header.extra_field_len);
    extrafield:=p64;
    longblockread(g.infile,pchar(extrafield)[0],header.extra_field_len,err);  {read in local header}
    if doserror=0 then begin
      extrafieldread:=0;
      while extrafieldread<header.extra_field_len do begin
        if (extrafield^.Tag=1) then begin
          {Note: the zip64 header is special: it's not a fixed size header,
           but contains only those fields which are set to $FFFF in the normal
           header! Fields which are not $FFFF are skipped, not left empty!}
          if (extrafield^.Size>=16) and (header.uncompress_size=-1) and (header.compress_size=-1) then begin
            g.uncompsize:=makecomp(extrafield^.orglow,extrafield^.orghi);
            g.compsize:=makecomp(extrafield^.complow,extrafield^.comphi);
          end else if (extrafield^.Size>=8) then begin {Only one of the two sizes!}
            gtemp:=makecomp(extrafield^.orglow,extrafield^.orghi);
            if (header.uncompress_size=-1) then
              g.uncompsize:=gtemp
            else if (header.compress_size=-1) then
              g.compsize:=gtemp;
          end;
        end else if (ziptype=99) and (extrafield^.Tag=$9901) and (extrafield^.Size>=7) then begin {AES}
          extrafield_aes:=paeslocalheader(extrafield);
          with extrafield_aes^ do begin
            if ((vendorversion=1) or (vendorversion=2)) and
               (vendorid[0]='A') and (vendorid[1]='E') then begin
              g.aes_strength:=aes_strength;
              ziptype:=zip_compression;
            end;
          end;
        end;
        if extrafield^.Size<=0 then break;
        inc(extrafieldread,extrafield^.Size+4);
        extrafield:=p64localheader(pchar(extrafield)+extrafield^.Size+4);
      end;
    end;
    freemem(p64,header.extra_field_len);
  end;
  compoffset:=compoffset+header.filename_len+header.extra_field_len+sizeof(tlocalheader);
  createsize:=g.uncompsize;
  originalcrc:=header.crc_32;
  if ((g.hufttype and 8)<>0) then begin {Size and crc at the end: Due to a bug this bit is set in DynaZip!}
    {Test for DynaZip bug:}
    if (ziptype<>0) then begin
      if (originalcrc=0) or (g.compsize=0) then begin
        g.compsize:=specialmaxcomp;           {Don't get a sudden zipeof!}
        g.uncompsize:=specialmaxcomp;
        createsize:=0;
        originalcrc:=0;
      end;
    end else if (originalsize<>-1) then begin
      {Special case: file is stored with size and CRC at the end -> use size from central dir header!}
      g.compsize:=originalsize;
      g.uncompsize:=originalsize;
      createsize:=originalsize;
    end;
  end;
  if ((1 shl ziptype) and GetSupportedMethods=0) and (ziptype<>98) then begin  {Not Supported!!!}
    result:=unzip_NotSupported;
    inuse:=false;
    exit;
  end;
  {$ifndef encryption}
  if (g.hufttype and 1)<>0 then begin {encrypted}
    result:=unzip_Encrypted;
    inuse:=false;
    exit;
  end;
  {$endif}
  if (g.hufttype and 1)<>0 then {encrypted}
    ziplastcrypted:=true;     {Needed to re-ask for password in case of an error!}

  compseek(g.infile,compoffset,0);
  allowdeletebad:=true;

  if not g_tomemory then begin
    if out_name<>nil then strcopy(outname,out_name);
    if attr and fadirectory<>0 then begin
      p:=strend(outname)-1;
      if (outname[0]<>#0) and (p[0]<>'\') then strcat(pchar(@p[1]),'\');
    end;
    isadir:=(outname[strlen(outname)-1]='\');

    flag:=copyflag_sequential;
    if (g.aes_strength>0) then
      flag:=flag or copyflag_notruncate;
    g.outfile:=CreateValidOutfile2(outname,isadir,createsize,flag);
    if isadir or (doserror<>0) or (g.outfile=-1{invalid file handle}) then begin
      if isadir then begin
        result:=unzip_Ok;
        longsetfattr(outname,attr);       {**************}
      end else result:=unzip_WriteErr;
      inuse:=false;
      exit
    end;
    allowdeletebad:=(g.aes_strength=0) or (compfilesize(g.outfile)=0);
  end else
    g.outfile:=invalid_handle_value;

  g.dlghandle:=hFileAction;
  g.dlgnotify:=cm_index;

  result:=DoRealUnpack(true,ziptype,originalcrc,offset,g,g_byteswritten,
    g_tomemory,g_crconly,g_outbuf,g_bufsize,g_testpassword);
  unpackhandle:=g.infile;     {Can change when disk was changed!}

  if strcomp(in_name,g.infilename)<>0 then   {Name changed because of Winzip split archive!}
    strlcopy(in_name,g.infilename,sizeof(tdirtype)-1);

  resCRC:=g.crc32Val;
  hFileAction:=g.dlghandle;

  if not g_tomemory then begin
    if timedate<>0 then begin
      {$ifdef wincmd}
      if unpack_flush then
        LongFlush(g.outfile);  {Write file completely before setting time stamp}
      {$endif}
      if (result=0) and (originalcrc=g.crc32val) then
        longsetftime(g.outfile,timedate); {set zipped time and date of outfile}
    end;
    longclose(g.outfile);
  end;
  if Result<>0 then begin
    if not g_tomemory and allowdeletebad then
    {$ifdef wincmd}if not packer_nodelete then{$endif}
      if (result=unzip_CRCErr){$ifdef wincmd} or not packer_openpartial{$endif} then
        longerase(outname,false);
  end else if (originalcrc<>g.crc32val) then begin
    result:=unzip_CRCErr;
    if not g_tomemory and allowdeletebad then
      {$ifdef wincmd}if not packer_nodelete then{$endif}
        longerase(outname,false);
  end else begin
    finalpercent:=100;       {100 percent}
    {$ifdef windows}
    if hFileAction<>0 then
      sendmessage(hFileAction,wm_command,cm_index,longint(@finalpercent));
    if finalpercent=$FFFF then begin
      Result:=unzip_userabort;
      if not g_tomemory then longerase(outname,false);
    end;
    {$endif}
    if (Result=0) and not g_tomemory then begin  {No user abort!}
      SetFileTimeAfterwardsIfNeeded(outname,timedate,g_byteswritten);
      longsetfattr(outname,attr);
    end;
  end;
  inuse:=false;
end;

function UnzipFile(in_name:pchar;out_name:pchar;attr:word;offset,originalsize:comp;
  var hFileAction:thandle;cm_index:integer;var unpackhandle:file):integer;
var byteswritten,crc:longint;
begin
  byteswritten:=0;
  crc:=0;
  Result:=UnzipFileInternal(in_name,out_name,attr,offset,originalsize,hFileAction,cm_index,
    byteswritten,false,false,nil,0,crc,unpackhandle,false);
end;

function unzipfiletomemory(in_name:pchar;out_buf:pchar;var buf_size:longint;offset,originalsize:comp;
  var hFileAction:thandle;cm_index:integer;var unpackhandle:file):integer;
var byteswritten,crc:longint;
begin
  byteswritten:=0;
  crc:=0;
  Result:=unzipfileInternal(in_name,nil,0,offset,originalsize,hFileAction,cm_index,
    byteswritten,true,false,out_buf,buf_size,crc,unpackhandle,false);
  if (Result=unzip_writeerr) then Result:=unzip_ok;  {kleinere Buffergrsse als ntig}
  buf_size:=byteswritten;
end;

function UnzipTestIntegrity(in_name:pchar;offset,originalsize:comp;
  var hFileAction:thandle;cm_index:integer;var crc:longint;var unpackhandle:file):integer;
var byteswritten:longint;
begin
  Result:=UnzipFileInternal(in_name,nil,0,offset,originalsize,hFileAction,cm_index,
    byteswritten,true,true,nil,0,crc,unpackhandle,false);
  if (Result=unzip_writeerr) then Result:=unzip_ok;  {kleinere Buffergrsse als ntig}
end;

function UnzipTestPassword(in_name:pchar;offset:comp;
  var hFileAction:thandle;cm_index:integer;var crc:longint;var unpackhandle:file):integer;
var byteswritten:longint;
begin
  Result:=UnzipFileInternal(in_name,nil,0,offset,-1,hFileAction,cm_index,
    byteswritten,true,true,nil,0,crc,unpackhandle,true);
  if (Result=unzip_writeerr) then Result:=unzip_ok;  {kleinere Buffergrsse als ntig}
end;

procedure CloseZipFileHandle(unpackhandle:file);
begin
  if unpackhandle<>INVALID_HANDLE_VALUE then
    CloseFile(unpackhandle);
end;


end.
